//  _
// | |_ ___  _   _ _ __ ___ _ __  _ __
// | __/ _ \| | | | '__/ __| '_ \| '_ \
// | || (_) | |_| | | | (__| |_) | |_) |
//  \__\___/ \__,_|_|(_)___| .__/| .__/
//                         |_|   |_|
// Author: Love Billenius <lovbi127@student.liu.se>

#include "Tour.h"
#include "Node.h"
#include "Point.h"
#include <iostream>

Tour::Tour() : m_startNode(nullptr) {}

Tour::~Tour() {
    if (m_startNode == nullptr)
        return;

    Node *node = m_startNode;
    if (node->next == m_startNode) {
        delete node;
        return;
    }

    do {
        Node *next = node->next;
        delete node;
        node = next;
    } while (node != m_startNode);
}

void Tour::show() const {
    if (m_startNode == nullptr)
        return;

    Node *node = m_startNode;
    do {
        Point &p = node->point;
        // I'm not using std::endl here to avoid flushing stdout for every
        // iteration.
        std::cout << '(' << p.x << ", " << p.y << ")\n";
        node = node->next;
    } while (node != m_startNode);

    // Finally, flush!
    std::flush(std::cout);
}

void Tour::draw(QGraphicsScene *scene) const {
    if (m_startNode == nullptr)
        return;

    Node *node = m_startNode;
    do {
        node->point.draw(scene);
        node->point.drawTo(node->next->point, scene);
        node = node->next;
    } while (node != m_startNode);
}

int Tour::size() const {
    if (m_startNode == nullptr)
        return 0;

    int count = 0;
    Node *node = m_startNode;
    do {
        count++;
        node = node->next;
    } while (node != m_startNode);

    return count;
}

double Tour::distance() const {
    if (m_startNode == nullptr)
        return 0;

    double totalDistance = 0;
    Node *node = m_startNode;
    do {
        totalDistance += node->point.distanceTo(node->next->point);
        node = node->next;
    } while (node != m_startNode);

    return totalDistance;
}

void Tour::insertNearest(Point p) {
    if (m_startNode == nullptr) {
        m_startNode = new Node(p);
        m_startNode->next = m_startNode; // Make it cirkular
        return;
    }
    if (m_startNode->next == nullptr) {
        auto newNode = new Node(p, m_startNode);
        m_startNode->next = newNode;
        return;
    }

    Node *nearest = m_startNode;
    double minDistance = Q_INFINITY;

    Node *node = m_startNode;
    do {
        double currentDistance = node->point.distanceTo(p);
        if (currentDistance < minDistance) {
            minDistance = currentDistance;
            nearest = node;
        }
        node = node->next;
    } while (node != m_startNode);

    auto *newNode = new Node(p, nearest->next);
    nearest->next = newNode;
}

enum class Orientation { COLLINEAR, CLOCKWISE, COUNTER_CLOCKWISE };

// I'm not going to pretend that I came up with this on my own.
// The description for the algorithms is found both in the
// linear algebra course, and on this wiki
// https://en.wikipedia.org/w/index.php?title=Line-line_intersection&oldid=1229564037

// Calculate the orientation of the triplet (p, q, r)
inline Orientation orientation(const Point &p, const Point &q, const Point &r) {
    int val = (q.y - p.y) * (r.x - q.x) - (q.x - p.x) * (r.y - q.y);

    if (val == 0)
        return Orientation::COLLINEAR;
    return (val > 0) ? Orientation::CLOCKWISE : Orientation::COUNTER_CLOCKWISE;
}
// True if q lies on the segment p-r
inline bool isOnSegment(const Point &p, const Point &q, const Point &r) {
    return (q.x <= std::max(p.x, r.x) && q.x >= std::min(p.x, r.x) &&
            q.y <= std::max(p.y, r.y) && q.y >= std::min(p.y, r.y));
}
// Check if the two line segments p1q1 and p2q2 intersect
bool pointsIntersect(const Point &p1, const Point &q1, const Point &p2,
                     const Point &q2) {
    Orientation o1 = orientation(p1, q1, p2);
    Orientation o2 = orientation(p1, q1, q2);
    Orientation o3 = orientation(p2, q2, p1);
    Orientation o4 = orientation(p2, q2, q1);

    // General case: line segments intersect if they have different orientations
    if (o1 != o2 && o3 != o4)
        return true;

    // Special cases
    // p1, q1, p2 are collinear and p2 lies on segment p1q1
    if (o1 == Orientation::COLLINEAR && isOnSegment(p1, p2, q1))
        return true;
    // p1, q1, q2 are collinear and q2 lies on segment p1q1
    if (o2 == Orientation::COLLINEAR && isOnSegment(p1, q2, q1))
        return true;
    // p2, q2, p1 are collinear and p1 lies on segment p2q2
    if (o3 == Orientation::COLLINEAR && isOnSegment(p2, p1, q2))
        return true;
    // p2, q2, q1 are collinear and q1 lies on segment p2q2
    if (o4 == Orientation::COLLINEAR && isOnSegment(p2, q1, q2))
        return true;

    return false;
}

bool Tour::pointsCross(Point &p1, Point &p2) {
    if (m_startNode == nullptr || m_startNode->next == m_startNode)
        return false;
    Node *n1 = m_startNode, *n2 = m_startNode->next;
    do {
        Point &q1 = n1->point, &q2 = n2->point;
        if (pointsIntersect(p1, p2, q1, q2))
            return true;

        n1 = n2;
        n2 = n2->next;
    } while (n1 != m_startNode);

    return false;
}

void Tour::insertSmallest(Point p) {
    if (m_startNode == nullptr) {
        m_startNode = new Node(p);
        m_startNode->next = m_startNode; // Make it cirkular
        return;
    }
    if (m_startNode->next == nullptr) {
        auto newNode = new Node(p, m_startNode);
        m_startNode->next = newNode;
        return;
    }

    Node *bestNode = m_startNode;
    double minIncrease = Q_INFINITY;

    Node *node = m_startNode;
    do {
        double increase = node->point.distanceTo(p) +
                          p.distanceTo(node->next->point) -
                          node->point.distanceTo(node->next->point);
        if (increase < minIncrease) {
            minIncrease = increase;
            bestNode = node;
        }
        node = node->next;
    } while (node != m_startNode);

    auto *newNode = new Node(p, bestNode->next);
    bestNode->next = newNode;
}
